use std::fmt::{Display, Formatter};
use std::str::FromStr;

use derive_more::{Deref, From, Into};
use rgb::{RGB8, RGBA8};
use serde_with::{DeserializeFromStr, SerializeDisplay};
use thiserror::Error;

#[derive(Debug, Error)]
#[error("The input value does not match the required format.")]
pub struct ParsingError {}

fn parse_rgb8_from_hex_str(s: &str) -> Result<RGB8, ParsingError> {
    let first_index = if s.starts_with('#') { 1 } else { 0 };
    if s.len() - first_index == 6 {
        let r = u8::from_str_radix(&s[first_index..(first_index + 2)], 16).map_err(|_| ParsingError {})?;
        let g = u8::from_str_radix(&s[(first_index + 2)..(first_index + 4)], 16).map_err(|_| ParsingError {})?;
        let b = u8::from_str_radix(&s[(first_index + 4)..(first_index + 6)], 16).map_err(|_| ParsingError {})?;

        return Ok(RGB8::new(r, g, b));
    }

    Err(ParsingError {})
}

fn fmt_rgb8_as_hex_string(v: &RGB8, f: &mut Formatter<'_>) -> std::fmt::Result {
    f.write_fmt(format_args!("#{:02x}{:02x}{:02x}", v.r, v.g, v.b))
}

fn parse_rgba8_from_hex_str(s: &str) -> Result<RGBA8, ParsingError> {
    let first_index = if s.starts_with('#') { 1 } else { 0 };
    if s.len() - first_index == 8 {
        let r = u8::from_str_radix(&s[first_index..(first_index + 2)], 16).map_err(|_| ParsingError {})?;
        let g = u8::from_str_radix(&s[(first_index + 2)..(first_index + 4)], 16).map_err(|_| ParsingError {})?;
        let b = u8::from_str_radix(&s[(first_index + 4)..(first_index + 6)], 16).map_err(|_| ParsingError {})?;
        let a = u8::from_str_radix(&s[(first_index + 6)..(first_index + 8)], 16).map_err(|_| ParsingError {})?;

        return Ok(RGBA8::new(r, g, b, a));
    }

    Err(ParsingError {})
}

fn fmt_rgba8_as_hex_string(v: &RGBA8, f: &mut Formatter<'_>) -> std::fmt::Result {
    f.write_fmt(format_args!("#{:02x}{:02x}{:02x}{:02x}", v.r, v.g, v.b, v.a))
}

fn parse_rgb8_with_optional_alpha_from_hex_str(s: &str, fallback_alpha: u8) -> Result<RGBA8, ParsingError> {
    // optionally +1 for the '#'
    match s.len() {
        6 | 7 => Ok(parse_rgb8_from_hex_str(s)?.alpha(fallback_alpha)),
        8 | 9 => Ok(parse_rgba8_from_hex_str(s)?),
        _ => Err(ParsingError {}),
    }
}

#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash, Deref, From, Into, SerializeDisplay, DeserializeFromStr)]
pub struct RGB8Wrapper(RGB8);

impl FromStr for RGB8Wrapper {
    type Err = ParsingError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        parse_rgb8_from_hex_str(s).map(RGB8Wrapper)
    }
}

impl Display for RGB8Wrapper {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        fmt_rgb8_as_hex_string(&self.0, f)
    }
}

#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash, Deref, From, Into, SerializeDisplay, DeserializeFromStr)]
#[repr(transparent)]
pub struct RGBA8Wrapper(pub RGBA8);

impl FromStr for RGBA8Wrapper {
    type Err = ParsingError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        parse_rgba8_from_hex_str(s).map(RGBA8Wrapper)
    }
}

impl Display for RGBA8Wrapper {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        fmt_rgba8_as_hex_string(&self.0, f)
    }
}

#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash, Deref, From, Into, SerializeDisplay, DeserializeFromStr)]
#[repr(transparent)]
pub struct RGB8WithOptionalA(RGBA8);

impl FromStr for RGB8WithOptionalA {
    type Err = ParsingError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        parse_rgb8_with_optional_alpha_from_hex_str(s, 0xff).map(RGB8WithOptionalA)
    }
}

impl Display for RGB8WithOptionalA {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        fmt_rgba8_as_hex_string(&self.0, f)
    }
}
