use std::collections::HashMap;
use std::ffi::OsString;
use std::path::Path;
use std::process::Stdio;
use std::sync::Arc;

use color_eyre::eyre::{eyre, WrapErr};
use color_eyre::Result;
use is_executable::IsExecutable;
use itertools::Itertools;
use log::warn;
use serde::{Deserialize, Serialize};
use tokio::io::{AsyncBufReadExt, AsyncWriteExt, BufReader};
use tokio::process::{ChildStdin, Command};
use tokio::sync::mpsc;

use deckster_shared::handler_communication::{
    HandlerCommand, HandlerEvent, HandlerInitializationError, HandlerInitializationResultMessage, InitialHandlerMessage,
};
use deckster_shared::path::{KeyPath, KnobPath};

#[derive(Debug, Serialize, Deserialize)]
pub struct KeyOrKnobHandlerConfig {
    pub host_id: Box<str>,
    pub name: Box<str>,
    pub config: Arc<toml::Table>,
}

pub async fn start(
    host_id: Box<str>,
    handlers_directory: &Path,
    key_configs: HashMap<KeyPath, KeyOrKnobHandlerConfig>,
    knob_configs: HashMap<KnobPath, KeyOrKnobHandlerConfig>,
    commands_sender: mpsc::Sender<HandlerCommand>,
    mut events_receiver: tokio::sync::broadcast::Receiver<HandlerEvent>,
) -> Result<()> {
    let handler_names: Vec<OsString> = std::fs::read_dir(handlers_directory)
        .wrap_err_with(|| format!("while reading the handlers directory: {}", handlers_directory.to_string_lossy()))?
        .filter_map(|entry| {
            if let Ok(entry) = entry {
                let path = entry.path();
                if path.is_executable() {
                    return Some(path.file_name().unwrap().to_os_string());
                }
            }

            None
        })
        .collect();

    let mut handler_stdin_by_name: HashMap<Box<str>, ChildStdin> = HashMap::with_capacity(handler_names.len());

    let mut handler_name_by_key_path: HashMap<KeyPath, Box<str>> = HashMap::new();
    let mut handler_config_by_key_path_by_handler_name: HashMap<Box<str>, HashMap<KeyPath, Arc<toml::Table>>> = HashMap::new();

    for (path, config) in key_configs {
        if config.host_id != host_id {
            continue;
        }

        handler_name_by_key_path.insert(path.clone(), config.name.clone());
        handler_config_by_key_path_by_handler_name
            .entry(config.name)
            .or_default()
            .insert(path, Arc::clone(&config.config));
    }

    let mut handler_name_by_knob_path: HashMap<KnobPath, Box<str>> = HashMap::new();
    let mut handler_config_by_knob_path_by_handler_name: HashMap<Box<str>, HashMap<KnobPath, Arc<toml::Table>>> = HashMap::new();

    for (path, config) in knob_configs {
        if config.host_id != host_id {
            continue;
        }

        handler_name_by_knob_path.insert(path.clone(), config.name.clone());
        handler_config_by_knob_path_by_handler_name
            .entry(config.name)
            .or_default()
            .insert(path, Arc::clone(&config.config));
    }

    for handler_name in handler_names {
        let handler_name = handler_name
            .into_string()
            .map_err(|_| eyre!("Command names must be valid Unicode."))?
            .into_boxed_str();

        let (key_configs, knob_configs) = match (
            handler_config_by_key_path_by_handler_name.remove(&handler_name),
            handler_config_by_knob_path_by_handler_name.remove(&handler_name),
        ) {
            (None, None) => {
                warn!("Handler '{handler_name}' is not used by any key or knob.");
                continue;
            }
            (a, b) => (a.unwrap_or_default(), b.unwrap_or_default()),
        };

        let mut command = Command::new(handlers_directory.join(handler_name.to_string()))
            .arg("deckster-run")
            .stdin(Stdio::piped())
            .stdout(Stdio::piped())
            .spawn()
            .wrap_err_with(|| format!("while spawning handler: {handler_name}"))?;

        let mut stdout_lines = BufReader::new(command.stdout.take().unwrap()).lines();
        let mut stdin = command.stdin.unwrap();

        let initial_handler_message = InitialHandlerMessage { key_configs, knob_configs };

        let serialized_message = serde_json::to_string(&initial_handler_message).unwrap().into_boxed_str().into_boxed_bytes();

        stdin.write_all(&serialized_message).await.unwrap();
        stdin.write_u8(b'\n').await.unwrap();
        stdin.flush().await.unwrap();

        let result_line = stdout_lines.next_line().await?.unwrap();
        let result: HandlerInitializationResultMessage = serde_json::from_str(&result_line)?;

        if let HandlerInitializationResultMessage::Error { error } = result {
            #[rustfmt::skip]
            if let HandlerInitializationError::InvalidConfig { supports_keys, supports_knobs, .. } = error {
                if !supports_keys && !initial_handler_message.key_configs.is_empty() {
                    return Err(eyre!(
                        "The '{handler_name}' handler does not support keys, but these keys tried to use it: {}",
                        initial_handler_message.key_configs.keys().map(|k| k.to_string()).join(", ")
                    ));
                } else if !supports_knobs && !initial_handler_message.knob_configs.is_empty() {
                    return Err(eyre!(
                        "The '{handler_name}' handler does not support knobs, but these knobs tried to use it: {}",
                        initial_handler_message.knob_configs.keys().map(|k| k.to_string()).join(", ")
                    ));
                }
            };

            return Err(eyre!("Starting the '{handler_name}' handler failed: {error}"));
        }

        let commands_sender = commands_sender.clone();

        tokio::spawn(async move {
            while let Ok(Some(line)) = stdout_lines.next_line().await {
                if line.starts_with('{') {
                    let command = serde_json::from_str::<HandlerCommand>(&line).unwrap();

                    commands_sender.send(command).await.unwrap();
                } else {
                    println!("{}", line);
                }
            }
        });

        handler_stdin_by_name.insert(handler_name, stdin);
    }

    if let Some((handler_name, config_by_key_path)) = handler_config_by_key_path_by_handler_name.drain().next() {
        return Err(eyre!(
            "There is no executable file named '{handler_name}' in the handlers directory but these keys have it set as their handler: {}",
            config_by_key_path.keys().join(", ")
        ));
    }

    if let Some((handler_name, config_by_knob_path)) = handler_config_by_knob_path_by_handler_name.drain().next() {
        return Err(eyre!(
            "There is no executable file named '{handler_name}' in the handlers directory but these knobs have it set as their handler: {}",
            config_by_knob_path.keys().join(", ")
        ));
    }

    tokio::spawn(async move {
        while let Ok(event) = events_receiver.recv().await {
            let handler_name = if let Some(n) = match &event {
                HandlerEvent::Key { path, .. } => handler_name_by_key_path.get(path),
                HandlerEvent::Knob { path, .. } => handler_name_by_knob_path.get(path),
            } {
                n
            } else {
                continue;
            };

            let handler_stdin = handler_stdin_by_name.get_mut(handler_name).expect("was already checked above");
            let serialized_event = serde_json::to_string(&event).unwrap().into_boxed_str().into_boxed_bytes();

            handler_stdin.write_all(&serialized_event).await.unwrap();
            handler_stdin.write_u8(b'\n').await.unwrap();
            handler_stdin.flush().await.unwrap();
        }
    });

    Ok(())
}
