use std::collections::HashMap;
use std::fmt::format;
use std::str::FromStr;
use std::time::Duration;

use rumqttc::{Event, Incoming, MqttOptions, QoS};
use serde::{Deserialize, Serialize};
use tokio::sync::{broadcast, mpsc};

use deckster_shared::handler_communication::{HandlerCommand, HandlerEvent};
use deckster_shared::path::{KeyPath, KeyPosition, KnobPath, KnobPosition};
use deckster_shared::style::{KeyStyle, KnobStyle};

use crate::handler_host::KeyOrKnobHandlerConfig;
use crate::model::config::MqttConfig;

#[derive(Debug, Serialize, Deserialize)]
pub struct HandlerHostsConfig {
    run_id: Box<str>,
    keys: HashMap<KeyPath, KeyOrKnobHandlerConfig>,
    knobs: HashMap<KnobPath, KeyOrKnobHandlerConfig>,
}

pub async fn start_coordinator_mqtt_client(
    config: &MqttConfig,
    key_configs: &HashMap<KeyPath, KeyOrKnobHandlerConfig>,
    knob_configs: &HashMap<KnobPath, KeyOrKnobHandlerConfig>,
    commands_sender: mpsc::Sender<HandlerCommand>,
    mut events_receiver: broadcast::Receiver<HandlerEvent>,
) {
    let topic_prefix = config.topic_prefix.to_owned();

    let mut options = MqttOptions::new(&config.topic_prefix, &config.host, config.port);
    options.set_keep_alive(Duration::from_secs(3));
    options.set_clean_session(false);

    if let Some(credentials) = &config.credentials {
        options.set_credentials(&credentials.username, &credentials.password);
    }

    let (client, mut event_loop) = rumqttc::AsyncClient::new(options, 32);

    client.subscribe(format!("{topic_prefix}/keys/+/+/style"), QoS::ExactlyOnce).await.unwrap();
    client.subscribe(format!("{topic_prefix}/knobs/+/+/style"), QoS::ExactlyOnce).await.unwrap();
    client.subscribe(format!("{topic_prefix}/knobs/+/+/value"), QoS::ExactlyOnce).await.unwrap();

    tokio::spawn({
        let topic_prefix = topic_prefix.clone();
        let client = client.clone();

        async move {
            while let Ok(event) = events_receiver.recv().await {
                match event {
                    HandlerEvent::Key { path, event } => {
                        client
                            .publish(
                                format!("{topic_prefix}/keys/{path}/events"),
                                QoS::ExactlyOnce,
                                false,
                                serde_json::to_string(&event).unwrap(),
                            )
                            .await
                            .unwrap();
                    }
                    HandlerEvent::Knob { path, event } => {
                        client
                            .publish(
                                format!("{topic_prefix}/knobs/{path}/events"),
                                QoS::ExactlyOnce,
                                false,
                                serde_json::to_string(&event).unwrap(),
                            )
                            .await
                            .unwrap();
                    }
                }
            }
        }
    });

    tokio::spawn({
        let topic_prefix = topic_prefix.clone();

        async move {
            while let Ok(event) = event_loop.poll().await {
                if let Event::Incoming(Incoming::Publish(event)) = event {
                    let segments = event.topic.strip_prefix(&topic_prefix).unwrap().split('/').skip(1).collect::<Vec<&str>>();
                    let page_id = segments[1];
                    let position = segments[2];
                    let property = segments[3];

                    match segments[0] {
                        "keys" => {
                            if property == "style" {
                                let value = serde_json::from_slice::<Option<KeyStyle>>(&event.payload).unwrap();

                                commands_sender
                                    .send(HandlerCommand::SetKeyStyle {
                                        path: KeyPath {
                                            page_id: page_id.to_owned(),
                                            position: KeyPosition::from_str(position).unwrap(),
                                        },
                                        value,
                                    })
                                    .await
                                    .unwrap();
                            }
                        }
                        "knobs" => {
                            let position = KnobPosition::from_str(position).unwrap();

                            match property {
                                "style" => {
                                    let value = serde_json::from_slice::<Option<KnobStyle>>(&event.payload).unwrap();

                                    commands_sender
                                        .send(HandlerCommand::SetKnobStyle {
                                            path: KnobPath {
                                                page_id: page_id.to_owned(),
                                                position,
                                            },
                                            value,
                                        })
                                        .await
                                        .unwrap();
                                }
                                "value" => {
                                    let value = serde_json::from_slice::<Option<f32>>(&event.payload).unwrap();

                                    commands_sender
                                        .send(HandlerCommand::SetKnobValue {
                                            path: KnobPath {
                                                page_id: page_id.to_owned(),
                                                position,
                                            },
                                            value,
                                        })
                                        .await
                                        .unwrap();
                                }
                                _ => {}
                            }
                        }
                        _ => {}
                    };
                }
            }
        }
    });

    for (path, config) in key_configs {
        client
            .publish(
                format!("{topic_prefix}/keys/{path}/config"),
                QoS::AtLeastOnce,
                true,
                serde_json::to_string(config).unwrap(),
            )
            .await
            .unwrap()
    }

    for (path, config) in knob_configs {
        client
            .publish(
                format!("{topic_prefix}/knobs/{path}/config"),
                QoS::AtLeastOnce,
                true,
                serde_json::to_string(config).unwrap(),
            )
            .await
            .unwrap()
    }
}

pub async fn start_handler_host_mqtt_client(
    config: &MqttConfig,
    handler_hosts_config_sender: mpsc::Sender<Option<HandlerHostsConfig>>,
    mut commands_receiver: mpsc::Receiver<HandlerCommand>,
    events_sender: broadcast::Sender<HandlerEvent>,
) {
    let topic_prefix = config.topic_prefix.to_owned();

    let mut options = MqttOptions::new(&config.topic_prefix, &config.host, config.port);
    options.set_keep_alive(Duration::from_secs(3));
    options.set_clean_session(false);

    if let Some(credentials) = &config.credentials {
        options.set_credentials(&credentials.username, &credentials.password);
    }

    let (client, mut event_loop) = rumqttc::AsyncClient::new(options, 32);

    client.subscribe(format!("{topic_prefix}/config"), QoS::ExactlyOnce).await.unwrap();
    client.subscribe(format!("{topic_prefix}/keys/+/+/events"), QoS::ExactlyOnce).await.unwrap();
    client.subscribe(format!("{topic_prefix}/knobs/+/+/events"), QoS::ExactlyOnce).await.unwrap();

    tokio::spawn({
        let topic_prefix = topic_prefix.clone();

        async move {
            while let Ok(event) = event_loop.poll().await {
                if let Event::Incoming(Incoming::Publish(event)) = event {
                    let topic_segments = event.topic.strip_prefix(&topic_prefix).unwrap().split('/').skip(1).collect::<Vec<&str>>();

                    if topic_segments[0] == "config" {
                        if let Ok(config) = serde_json::from_slice(&event.payload) {
                            handler_hosts_config_sender.send(config).await.unwrap();
                        } else {
                            log::error!("Could not deserialize the latest configuration from {}", event.topic);
                            handler_hosts_config_sender.send(None).await.unwrap();
                        };
                    } else {
                        let page_id = topic_segments[1];
                        let position = topic_segments[2];
                        let property = topic_segments[3];

                        match topic_segments[0] {
                            "keys" if property == "events" => {
                                if let Ok(event) = serde_json::from_slice(&event.payload) {
                                    events_sender
                                        .send(HandlerEvent::Key {
                                            path: KeyPath {
                                                page_id: page_id.to_owned(),
                                                position: KeyPosition::from_str(position).unwrap(),
                                            },
                                            event,
                                        })
                                        .unwrap();
                                } else {
                                    log::error!("Could not deserialize the latest event from {}", event.topic);
                                };
                            }
                            "knobs" if property == "events" => {
                                if let Ok(event) = serde_json::from_slice(&event.payload) {
                                    events_sender
                                        .send(HandlerEvent::Knob {
                                            path: KnobPath {
                                                page_id: page_id.to_owned(),
                                                position: KnobPosition::from_str(position).unwrap(),
                                            },
                                            event,
                                        })
                                        .unwrap();
                                } else {
                                    log::error!("Could not deserialize the latest event from {}", event.topic);
                                };
                            }
                            _ => {}
                        }
                    }
                }
            }
        }
    });

    tokio::spawn(async move {
        while let Some(command) = commands_receiver.recv().await {
            match command {
                HandlerCommand::SetActivePages { .. } => log::warn!("HandlerCommand::SetActivePages is not supported for remote handlers."),
                HandlerCommand::SetKeyStyle { path, value } => client
                    .publish(
                        format!("{topic_prefix}/keys/{path}/style"),
                        QoS::ExactlyOnce,
                        false,
                        serde_json::to_vec(&value).unwrap(),
                    )
                    .await
                    .unwrap(),
                HandlerCommand::SetKnobStyle { path, value } => client
                    .publish(
                        format!("{topic_prefix}/knobs/{path}/style"),
                        QoS::ExactlyOnce,
                        false,
                        serde_json::to_vec(&value).unwrap(),
                    )
                    .await
                    .unwrap(),
                HandlerCommand::SetKnobValue { path, value } => client
                    .publish(
                        format!("{topic_prefix}/knobs/{path}/value"),
                        QoS::ExactlyOnce,
                        false,
                        serde_json::to_vec(&value).unwrap(),
                    )
                    .await
                    .unwrap(),
            };
        }
    });
}
